/*GPL*START*
 * ass6510 - a very simple 6510 assembler
 * 
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include "tappconfig.h"
#include "asm.h"
#include "tfiletools.h"

#define OUTPUT_EXTENSION ".6510"

// history:
// 1998:
// 29 Dec 18:45 v0.0.0 started (old source take from ass6510.dos.c, 1994 V0.94d (whatever that means)) (dos sucks, linux rules)
// 29 Dec 23:00 v0.0.1 first usable version
// 30 Dec 01:30 v0.0.2 works fine, bugfixes
// 30 Dec 18:51 v0.0.3 bugfixes, english messages, disassembling illegals as BYT and bugfixes
// 31 Dec 01:20 v0.0.4 ALG (align) added
// 31 Dec 18:16 v0.0.5 bugfixes (output)
// 1999:
// 01 Jan 13:41 v0.0.6 dsb_detect in options, config file added (asm.cc: 1244 lines, this 198)
// 01 Jan 18:01 v0.0.7 make-module/make-plain added (233)
// 01 Jan 15:08 v0.0.8 better reverse assembling, foced non zero page added: LDA $0001 is three byte, BYT sequences
// 09 Jan 15:17 v0.0.9 BEC/ENC comment added
// 09 Feb 23:03 v0.1.0 better commandline semantics, help styling, moved to 'dist'


const char *options[] = {
   "#usage='Usage: %n [OPTIONS] <infile>\n\n"
     "this is a very simple 6510 assembler\nnormally <infile> is assembled into headerless 6510 object code",
     "#trailer='\n%n version %v *** (C) 1998-1999 by Johannes Overmann\ncomments, bugs and suggestions welcome: %e\n%gpl'",
   "#stopat--", // stop option scanning after a -- parameter
   "#remove--", // remove first -- from parameter list
   "#onlycl",
   
   // options
   "name=module             , type=switch, char=m, help='generate loadable program module (start address in first two bytes (lo/hi))', headline='assembler options:  (<infile> is treated as assembler source file)'",
   "name=warnings           , type=switch, char=w, help='enable assembler warnings'",
   "name=stop-on-errors     , type=switch, char=f, help='assembler stops after printing 20 error messages'",
   "name=output-file        , type=string, char=o, param=FILE, help='set name of the output file FILE (def: <infile>" OUTPUT_EXTENSION ")'",
   "name=symbol-file        , type=string, char=s, param=FILE, help='wrtite all symbols to file FILE'",   

   // disassemble
   "name=disassemble        , type=switch, char=d, help='print disassembled listing (standard format)', headline='disassembler operartions:  (<infile> is object code, -m ind. loadable module)'",
   "name=disass-source      , type=switch, char=S, help='print disassembled listing as assembler source'",
   "name=hexdump            , type=switch, char=D, help='print hexdump of object module (8 bytes per line)'",
   "name=wide-hexdump       , type=switch, char=W, help='print wide hexdump of object module (16 bytes per line)'",
   "name=opcode-list        , type=switch,       , help='print a list of all opcodes and their hex encoding'",
   "name=illegal            , type=switch,       , help='print illegal opcodes as \\'???\\' instead of \\'BYT $xx\\'', headline='disassemble options:'",
   "name=no-dsb             , type=switch,       , help='do not print long sequences of BRK ($00) as \\'DSB n\\''",
   "name=no-nop             , type=switch,       , help='do not print long sequences of NOP ($ea) as \\'NOP (xn)\\''",
   "name=start-address      , type=int   , char=O, param=ADR, help='set start address to ADR  (eg for non loadable modules)'",
   
   // module
   "name=make-module        , type=int, char=M, param=ADR, help='make <infile> a loadable module at start address ADR (by adding ADR as first two bytes)', headline=module operations:",
   "name=make-plain         , type=switch, char=P, help='make module <infile> a headerless object file\n(remove first two bytes)'",
      
   // common   
   "name=verbose            , type=switch, char=v, help=verbose execution, headline=common options:",
   "EOL" // end of list
};

bool verbose;
bool warnings_enabled;
bool stop_on_errors;
bool loadable_module;
bool no_dsb;
bool no_nop;
bool dis_source;
bool illegal;
FILE *ifile;



static void load_obj(FILE *f, const char *fname) {
   int len = flen(fname);
   if(loadable_module) {
      len -= 2;
      // read base address      
      uchar c;
      c = fgetc(f);
      base |= c;
      c = fgetc(f);
      base |= ushort(c)<<8;
      if(verbose && (!dis_source))
	printf("loadable module\n");
   } else {
      base = tApp->getInt("start-address");
   }
   if(verbose && (!dis_source))
     printf("loading %d bytes of object code to base $%04X\n", len, base);
   fread(mem, 1, len, f);
   amem = base + len;     
}


static void save_obj(const char *fname) {
      // open output file
      FILE *ofile = fopen(fname, "w");
      if(ofile == 0)
	userError("can't open output file '%s' for writing!\n", fname);
      if(verbose) 
	printf("output file '%s' successfully opened for writing\n", fname);
      
      // write out file
      if(loadable_module) { // put start address (lo/hi) for loadable module
	 fputc(base&255, ofile);
	 fputc(base>>8, ofile);
	 if(verbose) 
	   printf("writing base address $%04X for loadable module\n", base);
      }
      if(verbose) 
	printf("writing %d bytes to file '%s'\n", amem-base, fname);
      fwrite(mem, 1, amem-base, ofile); // write object code   
      fclose(ofile);   
}



int main(int argc, char *argv[]) {
   // get parameters
   TAppConfig ac(options, "options", argc, argv, "ASS6510OPT", "ass6510rc", VERSION);
   
   // setup
   verbose             = ac("verbose");   
   warnings_enabled    = ac("warnings");   
   if(ac("stop-on-errors")) 
     max_err = 20;
   loadable_module     = ac("module");
   no_dsb              = ac("no-dsb");
   no_dsb              = ac("no-nop");
   illegal             = ac("illegal");
   dis_source          = ac("disass-source");
   TString infname;
   TString outfname;
   memset(mem, 0, max_mem); 

   if(!ac("opcode-list")) {
      // get infile/outfile
      if(ac.numParam()!=1) 
	userError("need exactly one input source file! (%d given)\n", ac.numParam());
      infname = ac.param(0);
      if(ac.getString("output-file")) outfname = ac.getString("output-file");
      else outfname = infname + OUTPUT_EXTENSION;

      // open if/outfile
      ifile = fopen(infname, "r");
      if(ifile == 0)
	userError("can't open input file '%s' for reading!\n", *infname);
      if(verbose && (!dis_source)) 
	printf("input file '%s' successfully opened for reading\n", *infname);
   }
      
   if(ac("disassemble")||ac("disass-source")) {
      
      // disassemble
      load_obj(ifile, infname);
      disassemble(false);
      
   } else if(ac("opcode-list")) {
      
      // opcode list
      mem[1] = 0xaa;
      mem[2] = 0xbb;
      base = 0x1000;
      amem = base+1;      
      for(int i=0; i<256; i++) {
	 mem[0] = i;
	 disassemble(true);
      }
      
   } else if(ac("hexdump")) {
      
      // hexdump
      load_obj(ifile, infname);
      hexDump(8);
      
   } else if(ac("wide-hexdump")) {
      
      // wide hexdump
      load_obj(ifile, infname);
      hexDump(16);
      
   } else if(ac.wasSetIn("make-module")!=DEFAULT) {
      
      // make a loadable module      
      load_obj(ifile, infname);
      fclose(ifile);
      int len = amem-base;
      loadable_module = true;
      base = ac.getInt("make-module") & 65535;
      amem = base + len;
      save_obj(infname);
      
   } else if(ac("make-plain")) {
      
      // remove first two bytes
      loadable_module = true;
      load_obj(ifile, infname);
      fclose(ifile);
      loadable_module = false;
      save_obj(infname);
      
   } else {
      // assemble

      // init assembler
      if(verbose) printf("\033[33;01m");
      if(!initAss()) {
	 printf("                                  \033[00m\n");
	 userError("no 'ASS'-command found in input file (no assembler source?)\n");
      }

      // assemble
      while(assLine());
      printUnused();
      if(verbose) printf("                                  \033[00m\r");
      fclose(ifile);
      
      // save object file
      save_obj(outfname);
      
      if(ac.getString("symbol-file")) {
	 
	 // write symbol file
	 FILE *f = fopen(ac.getString("symbol-file"), "w");
	 if(f==0) 
	   userError("can't open symbol file '%s' for writing!\n", *ac.getString("symbol-file"));
	 if(verbose) 
	   printf("writing symbols to file '%s'\n", *ac.getString("symbol-file"));
	 printSymbols(f);
	 fclose(f);
	 
      }
      
      if(verbose)
	printStats();
      
   } // if(assemble) ...
         
   // normal exit
   return 0;
}


